import os
from pathlib import Path
import json
import logging
import pprint
from datetime import datetime
import sys 
import tempfile

import ocrmypdf
from markitdown import MarkItDown, FileConversionException, UnsupportedFormatException

from .paperless import Paperless
from .model import Document, JsonItem, CustomField
from .tagger import Tagger
from .config import get_config
from .model import Document,CustomField, data_load, event_dict, document_dict


PROMPT_DOCUMENT = """
Analizza il testo fornito e genera un output in formato JSON.
Assicurati che l'output sia preciso e ben strutturato. Restituisci solo l'oggetto JSON, senza ulteriori testi o spiegazioni.
"""

PROMPT_DATE = """
Extract date from the attached row
Dates appear in various formats and could be only the year
generate a well formatted JSON file, with this structure
{"row": text of row, "date": [extracted date  in ISO format]}
If only the year is found, use only the year, do not invent
Return only the JSON object, without any further text or explanation
"""

config = get_config()

log = logging.getLogger(__name__)

def do_markitdown(filepath: str):
    md = MarkItDown()
    log.info(f"Conversione del file {filepath}")

    try:
        result = md.convert(filepath)
    except FileConversionException as e:
        log.error(f"La conversione del file {filepath} ha prodotto l'errore: {e}.")
        return None 
    except UnsupportedFormatException as e:
        log.error(f"Formato del file {filepath} non supportato.")
        return None
    except:
        log.error(f"La conversione del file {filepath} ha prodotto un errore non gestito")
        return None
    return result


def check_if_scanned_pdf(file: str, target_file: str):
    """
    Controlla se il pdf è da scansione, supponendo che lo sia se contiene del testo e lo passa all'ocr.
    """
  
    if os.stat(target_file).st_size == 0:     
        with tempfile.NamedTemporaryFile(mode='w+', delete=False) as tmp_file:
            ocrmypdf.ocr(input_file=file, output_file=tmp_file.name, language="ita")
            tmp_path = tmp_file.name
            log.debug(f"Creato file temporaneo OCR {tmp_path}")
            result = do_markitdown(tmp_path)
            if result:
                with open(target_file, 'w') as f:
                    f.write(result.text_content)
                    log.debug(f"File markdown per {file} creato.")   
        if tmp_path:
            os.remove(tmp_path)


def upload_document(file_list):
    tagger = Tagger()
    result = tagger.traverse(file_list)

    # for file in result:
    #     root_name, extension = os.path.splitext(file)
    #     # extension = os.path.splitext(file)[1].lower()
    #     extension = os.path.splitext(file)[1].lower()
    #     if extension.lower() == ".json":
    documents = [file for file in result if os.path.splitext(file)[1].lower() in config.extension]
        
    correspondents = Paperless.get_correspondents()
    document_types = Paperless.get_document_types()

    # E' un test
    # Prendo i pdf nella directory. Se c'e' un json corrispondente procedo
    #for document in list(source_path.glob("*.pdf")):
    for document in documents:
        json_source_path = os.path.splitext(document)[0] + ".json"
        if not os.path.exists(json_source_path):
            continue
        
        paper = Paperless()
        paperless_custom_fields = paper.get_custom_fields()

        data = paper.import_from_json(json_source_path)
        data_evento = data['data_evento']
        data['data_evento'] = datetime.fromisoformat(data_evento)
        
        documento = Document(
            persone_citate= data['persone_citate'] if data.get('persone_citate') else [],
            data_evento=data['data_evento'],
            relatori=data['relatori'] if data.get('relatori') else [],
            macroargomenti_trattati=data['macroargomenti_trattati'] if data.get('macroargomenti_trattati') else [],
            tipo_documento=data['tipo_documento']
        )

        # Senza estensione
        title = os.path.basename(document)
        title = os.path.splitext(title)[0]
        payload = {
            "title": title,
            "created": data_evento,

        }

        # Estraggo evento
        radice = os.path.dirname(config.files[0])
        directory_evento = Path(document).relative_to(radice).parts[0]
        evento = event_dict.get(directory_evento, "Sconosciuto")        
        
        # Correspondent
        for correspondent  in correspondents:
            if correspondent.name == directory_evento:
                payload["correspondent"] = correspondent.id
                break

        # Document Typedocument_dict
        # # TODO se non dovesse esserci il tipo di documenta (???) assegno un tipo di default (Comunicato stampa)
        dc = document_dict.get(documento.get("tipo_documento", "cs"))
        for document_type  in document_types:
            if document_type.name == dc:
                payload["document_type"] = document_type.id
                break
       
        # Anno come tag
        tags = Paperless.get_tags()
        
        anno_evento = str(documento.get("data_evento").year) if documento.get("data_evento") else None
        if anno_evento:
            tag_id = None
            for tag_anno in tags:
                if tag_anno == anno_evento:
                    tag_id = tags[tag_anno]
                    break
            if not tag_id:
                tag_id = Paperless.post_tag(anno_evento)
            if tag_id:
                payload["tags"] = [tag_id]
        pprint.pprint(payload)


        # Custom Fields
        cf = CustomField(documento, evento, paperless_custom_fields)
        cf_obj = cf.custom_field_object()

        task_id = paper.post_document(document, payload)
        result = paper.taskid_check(task_id)
        
        # result = {
        #     "status":"SUCCESS",
        #     "related_document": 8923
        # }
        
        if result and result.get("status") == "SUCCESS":
            # mando i custom fields usando il document id
            response = paper.update_custom_fields(result["related_document"], cf_obj)
        else:
            error_msg = "Fallito caricamento documento {documento}. "
            if result and result.get("status") == "FAILURE":
                error_msg += result.get("result")
            log.error(error_msg)

def do_stats(file_list):
    tagger = Tagger()
    result = tagger.traverse(file_list)
    
    custom_fields =[]
    json_files = []
    tokens = 0

    document_status = tagger.document_status(file_list)
    if len(document_status) == 0:
        log.info("Non ci sono documenti convertiti o elaborati.")
        return

    for file in result:
        extension = os.path.splitext(file)[1].lower()
        if extension == ".json":
            documento = data_load(file)
            # Salto il global.json
            if documento.get("tipo") and documento.get("tipo") == "riepilogo":
                continue
            # Salto i files corti, classificati come img
            if documento.get("tipo_documento") and documento.get("tipo_documento") == "img":
                continue
            radice = os.path.dirname(config.files[0])
            directory_evento = Path(file).relative_to(radice).parts[0]  

            #directory_evento = Path(file).relative_to(config.root).parts[0]
            
            evento = event_dict.get(directory_evento, "Sconosciuto")
            cf = CustomField(documento, evento)
            custom_fields.append(cf)
            json_files.append(file)
        elif extension == ".md":
            with open(file, "r") as f:
                testo_da_elaborare = f.read()
                tokens += len(testo_da_elaborare.split())
   
    
    # I token si stima vadano divisi per 4
    tokens = tokens / 4

    # 1 - Statistiche sui files nel percorso indicato
    #     Files processati/Files non processati
    #     Numero files catalogati per estensione
    # # # elenco file convertiti e non
    converted = [item["path"] for item in document_status.values() if item["converted"]]
    not_converted =[item["path"] for item in document_status.values() if not item["converted"]]

    converted_ratio = len(converted) / len(document_status) if len(document_status) > 0 else -1
    assert converted_ratio >= 0
    # # # # ratio di  not_converted e  converted 
    not_converted_ratio = 1 -  converted_ratio
     
    if config.debug:   
        print(f"**** Elenco file convertiti")
        for i in converted:
            print( i )   
        print(f"**** Elenco file non convertiti")
        for i in not_converted:
            print( i)

    print(f"Totale files: {len(document_status)}")
    print(f"Totale files convertiti {len(converted)} ({converted_ratio * 100:.2f}%)")
    print(f"Totale files non convertiti {len(not_converted)} ({not_converted_ratio * 100:.2f}%)")

    # # # elenco file processati e non 
    processed = [item["path"] for item in document_status.values() if item["processed"]]
    not_processed =[item["path"] for item in document_status.values() if not item["processed"]]
    assert len(document_status) != 0
    processed_ratio = len(processed) / len(document_status) if len(document_status) > 0 else -1
    assert processed_ratio >= 0
    # # # # ratio di  not_processed e  processed 
    not_processed_ratio = 1 -  processed_ratio
    
    if config.debug:
        print(f"**** Elenco file processati")
        for i in processed:
            print( i )   
        print(f"**** Elenco file non processati")
        for i in not_processed:
            print( i)

    print(f"Totale files: {len(document_status)}")
    print(f"Tolale files processati {len(processed)} ({processed_ratio * 100:.2f}%)")
    print(f"Tolale files non processati {len(not_processed)} ({not_processed_ratio * 100:.2f}%)")

    # Conteggio token
    print(f"**** Token richiesti: {tokens}")

    # 2 - Elenco del contenuto dei custom fields ordinato per 
    #     cardinalità decrescente
    #     elenco in property_to_name_map (aggiungere anche tipo_documento)

    cf_stat = tagger.cf_stats(custom_fields)
 
    # intanto lo metto qui ,provvisoriamente
    radice = os.path.dirname(config.files[0])  
    stats_file = os.path.join(radice, "global.json")
    
    for k,v in cf_stat.items():
        print (f"******** {k} ******")
        for n in v:
            print(n[0], ":", n[1])

    cf_stat["tipo"] = "riepilogo"
    cf_stat["token"] = tokens
    with open(stats_file, 'w') as f:
        json.dump(cf_stat, f)


# Non sta bene mel main
def convert(source_path: str, extensions: list):
    """ Converte files in markdown.
    
    Esplora ricorsivamente un percorso e converte tutti quelli che hanno estensione ammessa.

    Args:
        source_path (str): la stringa contentente il path del file
        extensione  (list): la lista delle estensioni consentite 

    Returns:
        None

    """
    tagger = Tagger()
    files = tagger.traverse(source_path)
    
     
    for file in files:
        base, extension = os.path.splitext(file)
        
        if extension.lower() in extensions:
            target_file = base + ".md"
            if os.path.exists(target_file):
                 log.info(f"File markdown per {file} presente. Nessuna conversione")
                 continue

            # markitdown
            # md = MarkItDown()
            # log.info(f"Conversione del file {file}")
            # try:
            #     result = md.convert(file)
            # except FileConversionException as e:
            #     log.error(f"La conversione del file {file} ha prodotto l'errore: {e}.")
            #     continue
            # except UnsupportedFormatException as e:
            #     log.error(f"Formato del file {file} non supportato.")
            #     continue
            # except:
            #     log.error(f"La conversione del file {file} ha prodotto un errore non gestito")
            #     continue

            #######
            result = do_markitdown(file)
            if result:
                with open(target_file, 'w') as f:
                    f.write(result.text_content)
                    log.info(f"File markdown per {file} creato.")
                    # Controllo se estensione è pdf e dimensione md == 0 (probabile grafica) 
                    # in questo caso lo passo all'ocr

                if extension == ".pdf":
                    check_if_scanned_pdf(file, target_file)

                
def ai(path):
    tagger = Tagger()
    all_files = tagger.traverse(path)

    MD_EXT = ".md"

    for source_file in all_files:
        # Processa solo i documenti con estensione opportuna
        if not os.path.splitext(source_file)[1] in config.extension:
            continue

        target_file = os.path.splitext(source_file)[0] + MD_EXT
        if not os.path.exists(target_file):
            continue

        with open(target_file, "r") as f:
            testo_da_elaborare = f.read()
        
        tag_file = os.path.join(target_file, os.path.splitext(target_file)[0] + ".json")
        if os.path.exists(tag_file):
            log.debug(f"File JSON per {target_file} presente. Elaborazione dati non richiesta")
            continue

        # Conta token. Se sotto un certo limite, non elaborare
        tokens = testo_da_elaborare.split()
        if len(tokens) < 30:
            tagged_result = {
                "tipo_documento": "img"
            }
            data_estratta = ""
        else:
            tagged_result = tagger.analyze(PROMPT_DOCUMENT, testo_da_elaborare, template=Document)

            # Controllo data mandando  all'ai il nome del file
            # [{"row": nomefile, "date: []"}]
            # Un po' debole. 
            log.debug(f"Ricerca data in {os.path.basename(source_file)}")
            response = tagger.analyze(PROMPT_DATE, os.path.basename(source_file))
            if isinstance(response, list):
                response = response[0]
            log.debug(f"risultato ricerca data {response}")
            date_field = response.get("date",list())
            data_estratta = date_field[0] if len(date_field) > 0 else ""

        if not tagger.is_acceptable_data(data_estratta):
            log.debug(f"La data esttratta dal nome del file {os.path.basename(source_file)} non è completa ({data_estratta})")
            tagged_result["data_evento"] = tagger.get_last_change(source_file)
            log.debug(f"Uso data ultima modifica: {tagged_result["data_evento"]}")
        else:
            tagged_result["data_evento"] = data_estratta

        with open(tag_file, 'w') as f:
            json.dump(tagged_result, f)
            log.debug(f"Processato {target_file}. Creato corrispondente JSON")
        # FIXME Provvisorio durante le prove per superare i limiti di token per minuto    
        import time
        time.sleep(3)


def main():
    if not len(config.files):
        log.error(f"E' richiesta una directory di documenti.")
        sys.exit(-1)
    match  config.task:
        case "convert":
            convert(config.files[0], config.extension)
        case "evaluate":
            ai(config.files[0])
        case "report":
            do_stats(config.files[0])
        case "upload":
            upload_document(config.files[0])
        
