from typing import List
from typing_extensions import Annotated, TypedDict
from datetime import datetime
from dataclasses import dataclass
import json

from .config import get_config

config = get_config()

MAX_LENGTH = 128

event_dict = {
    "ARTVERONA": "art",
    "CORPORATE": "gen",
    "EOS": "eos",
    "FIERACAVALLI": "fic",
    "FIERAGRICOLA": "agr",
    "JOB&ORIENTA": "job",
    "LETEXPO": "tra",
    "MARMOMAC": "mdm",
    "MCTER EXPO": "eio",
    "MINERAL MAGGIO": "min",
    "MINERAL NOVEMBRE": "geo",
    "MODEL EXPO": "mod",
    "MOTORBIKE": "mbe",
    "PAVING": "pav",
    "PROGETTO FUOCO": "fuo",
    "SAMOTER": "sam",
    "SOL2EXPO": "sag",
    "SPORT EXPO": "spo",
    "VINITALY": "vir",
    "AND THE CITY": "vac",
    "VITA IN CAMPAGNA": "vca",
    "WINE2WINE": "w2w",
}

document_dict = {
    "cs": "Comunicato Stampa",
    "sp": "Speech",
    "cst": "Cartella Stampa",
    "red": "Redazionali",
    "int": "Intervista",
    "prg": "Programma",
    "scal": "Scaletta",
    "rep": "Report",
    "soc": "Social",
    "std": "Save The Date",
    "inv": "Inviti Stampa",
    "cs3": "Comunicati Stampa Terzi",
    "pk": "Press Kit",
    "pr": "Press Release",
    "img": "Logo"
}

# Questo è il documento lato Gemini, ovvero dal documento originale al risultato dell'estrazione dati
# ed è usato per serializzare il .json 
class Document(TypedDict):
    persone_citate: Annotated[List[str], ..., "Elenco di nomi di persone citate (es. artisti, curatori, figure rilevanti)."]
    data_evento: Annotated[datetime, ...,"Data di riferimento dell'evento (formato YYYY-MM-DD)."]
    relatori: Annotated[List[str], ..., "Elenco di nomi dei relatori, o chi commenta, se presenti nel testo."]
    macroargomenti_trattati: Annotated[List[str], ..., "Elenco dei temi principali trattati (es. sostenibilità, arte, ambiente)."]
    tipo_documento: Annotated[str,"non_classificato", """Classifica il documento scegliendo tra i seguenti codici:
        - cs: Comunicato stampa (testo per comunicare contenuti alla stampa).
        - cst: Cartella di comunicati (riepilogo di comunicati stampa).
        - red: Redazionali (articoli di giornale inviati a riviste).
        - sp: Discorso.
        - scal: Scaletta (programma dettagliato con orari).
        - rep: Report (resoconto di riunione).
        - soc: Social (testi per piattaforme social).
        - std: Inviti.
        - cs3: Comunicati stampa (specifica variante)."""]
    

def data_load(filename):
    document = Document()
    
    with open(filename, 'r', encoding='utf-8') as file:
        data = json.load(file)
    
    for key in data:
        document[key] = data[key]
    document["data_evento"] = datetime.fromisoformat(document.get("data_evento")) \
                              if document.get("data_evento") \
                              else None
    return document


# CUSTOM FIELDS MAPPING
# Evitiamo sorprese in caso di modifiche in Paperless
# chiave: nome attributo classe Document (es self.anno)
# valore: nome Custom Field in Paperless (es Anno), QUI scritto in MINUSCOLO
property_to_name_map = {
            "persone_citate": "persone citate",
            "argomenti": "argomenti",
            "anno": "anno",
            "tipo_evento": "tipo evento",
            "persone_evidenziate": "persone evidenziate",
            "relatori": "relatori",
            "filename" : "nome file"
        }


# Correspondent Come lo ricevo da Paperless
@dataclass
class Correspondent:
    id: int
    slug: str
    name: str
    match: str
    matching_algorithm: int
    is_insensitive: bool
    document_count: int
    owner: int
    user_can_change: bool 
    

# Document Type Come lo ricevo da Paperless
@dataclass
class DocumentType:
      id: int
      slug: str
      name: str
      match: str
      matching_algorithm: int
      is_insensitive: bool
      document_count: int
      owner: int
      user_can_change: bool


# Custom Field Come lo ricevo da Paperless
@dataclass
class JsonItem:
    id: int
    name: str
    data_type: str
    extra_data: dict
    document_count: int 

# Custom Field come lo mando a Paperless
@dataclass
class FinalItem:
    id: int
    value: str

# Sono i "Campi personalizzati" di paperless. La maggiorparte di essi la si ritrova nel .json
# Uso la classe Document per caricare i dati dal file su disco e poi li estraggo
class CustomField:
    def __init__(self, documento: Document, tipo_evento: str, paperless_cf=[]): 
        self.persone_citate = documento.get("persone_citate",[])
        self.argomenti = documento.get("macroargomenti_trattati",[])
        self.anno = documento.get("data_evento").year if documento.get("data_evento") else None
        self.tipo_evento = tipo_evento
        self.relatori = documento.get("relatori", [])
        self.paperless_custom_fields = [JsonItem(**item) for item in paperless_cf]
        
        # Persone citate che hanno particolare evidenza
        self.persone_evidenziate = set(config.person) & {x.title() for x in self.persone_citate}
        # Nome del file secondo convenzioni VeronaFIere
        #<tipo documento>-<data>-<elenco persone evidenziate>
        data_evento = "_".join([
            str(documento.get("data_evento").year),
            str(documento.get("data_evento").month),
            str(documento.get("data_evento").day)])
        
        filename = [
            documento.get("tipo_documento")
        ]
        persone_evidenziate = "_".join(self.persone_evidenziate)
        persone_evidenziate = persone_evidenziate.lower().replace(" ", "_")
        assert data_evento != None
        assert len(filename) != 0
        
        filename.append(data_evento)
        if len(persone_evidenziate) > 0: 
            filename.append(persone_evidenziate)
        self.filename = "-".join(filename)
        
            
    # Faccio il mapping tra la proprietà (es self.persone_citate) e l'id del 
    # corrisondente custom field in paperless per la HTTP patch
    #Es extracted_ids = {'argomenti': 2, 'persone_citate': 3}
    def extract_ids_from_json(self, json_data: List[JsonItem]) -> dict:
        # Dizionario per memorizzare gli id estratti
        extracted_ids = {}

        # Scansiona il JSON per trovare gli id corrispondenti
        #property_name=="tipo_evento"  Proprietà Classe CustomField
        #json_name=="tipo evento" 
        #item.name== "Persone Citate" Nome Proprietà in paperless associata ad un id che devo ricavare
        for item in json_data:
            for property_name, json_name in property_to_name_map.items():
                if item.name.lower() == json_name:
                    extracted_ids[property_name] = item.id

        return extracted_ids   
    
    def custom_field_to_dict(self):
        return {attr: value for attr, value in self.__dict__.items() if attr in property_to_name_map}

    # Creo l'oggetto che usetò nella HTTP patch per caricare il custom field
    # FinalItem(id=2, value=['vini', 'fiere', 'export'])]
    def custom_field_object(self) -> List[FinalItem]:
        # Estrai gli id dal JSON
        extracted_ids = self.extract_ids_from_json(self.paperless_custom_fields)

        # Bozza... Dovrebbe adattare quello che arriva da JSON al tipo
        # assegnato al Custom Field in Paperles
        # In realtà dovrei controllare in JsonItem il tipo richiesto
        def format_field(value):
                if isinstance(value, list) or isinstance(value, set):
                    # Controllo il limite di lunghezza
                    length = 0
                    temp_list = []
                    for i in value:
                        length = length + len(i) + (1 if temp_list else 0)
                        if length > MAX_LENGTH:
                            break
                        temp_list.append(i)
                    field = ",".join(temp_list)
                elif  isinstance(value, int):
                    field = value
                else:
                    field = str(value)[:MAX_LENGTH]
                return field
 
        # Costruisci l'oggetto finale utilizzando i dataclass
        # in questo modo posso che custom field inviare , in caso di successive modfiche
        custom_field_object = [
            FinalItem(id=extracted_ids.get("persone_citate"), value=format_field(self.persone_citate)),
            FinalItem(id=extracted_ids.get("argomenti"), value=format_field(self.argomenti)),
            FinalItem(id=extracted_ids.get("anno"), value=format_field(self.anno)),
            FinalItem(id=extracted_ids.get("tipo_evento"), value=format_field(self.tipo_evento)),
            FinalItem(id=extracted_ids.get("relatori"), value=format_field(self.relatori)),
            FinalItem(id=extracted_ids.get("persone_evidenziate"), value=format_field(self.persone_evidenziate)),
            FinalItem(id=extracted_ids.get("filename"), value=format_field(self.filename))
        ]
        
        return custom_field_object
    
    def __str__(self):
        return f"Persone citate: {self.persone_citate}\n" + \
               f"Argomenti: {self.argomenti}\n" + \
               f"Anno evento: {self.anno}\n" + \
               f"Tipo Evento: {self.tipo_evento}\n" + \
               f"Persone Evidenziate: {self.persone_evidenziate}\n" + \
               f"Relatori: {self.relatori}\n\n"
    
