import glob
import hashlib
import os
from collections import defaultdict, Counter
import logging
from langchain_core.messages import HumanMessage, SystemMessage
import re
import datetime
from typing import Optional, Dict, Any, Union
import json
import traceback

from .ai import Ai
from .model import Document
from .config import get_config

config = get_config()

log = logging.getLogger(__name__)

extensions = [".pdf", ".odt", ".txt", ".pptx", ".doc", ".xlsx", ".docx"]

class Tagger:
    def __init__(self):
        self.ai = Ai()

    def analyze(self, prompt: str, 
                payload: str, 
                system_message: str = "Sei un sistema di indicizzazione per documenti.",
                template: Any=None) -> Union[str, object]:
        """
        Richiede un'analisi ad un LLM.

        L'analisi può essere di tipo struttura, se viene fornito una classe template oppure
        non strutturato.

        Args:
            json_string (str): La stringa contenente il JSON formattato con ```json ... ```.
            prompt (str): il prompt da fornire al motore LLM
            payload (str); il testo sul quale effettuare l'analisi
            system_message (str): lo scopo assegnato alla sessione
            template (Any): un template per dati strutturati

        Returns:
            str: la risposta ottenuta dal motore llm

        """        
 
        messages = [
            SystemMessage(system_message),
            HumanMessage(
                content=[
                    {"type": "text", "text": prompt},
                    {"type": "text", "text": payload}
                ]
            ),
        ]
        
        if template:
            return self.ai.structured_invoke(template, messages)
        else:
            response = self.ai.invoke(messages)
            return self.parse_json(response.content)
        

    def traverse(self, path: str):
        #path = '/mnt/disk/Verona*'
        # path = '/etc/resolv.conf'
        result = []
         
        if isinstance(path, str) and "*" in path:
            result.extend(os.path.abspath(f) for f in glob.iglob(path, recursive=True))
        else: 
            result.append(os.path.abspath(path))
        valid_extensions = tuple(config.extension + [".md", ".json"])
        filtered_files = [f for f in result if os.path.splitext(f)[1].lower() in valid_extensions]
        return filtered_files
    

    def document_status(self, file_list):
        """Effettua l'attraversamento di un path.
        
        Ritorna un oggetto con lo stato di un documento in relazione ai
        files presenti che  hanno lo stesso path, ma differiscono nell'estensione
        """
        # gtagger --debug "**"
        # gtagger --debug "**.py"

        # {
        #     "hash_path" : {
        #         "path": "path completo senza estensione",
        #         "type": "pdf",
        #         "converted": True,
        #         "processed": False 
            
        #     }
        
        # }
        files = self.traverse(file_list)
        
        dir_content = {}

        for filename in files:
            
            extension = os.path.splitext(filename)[1].lower()
            # if extension not in valid_extensions:
            #     continue
            if os.path.basename(filename) == "global.json":
                continue
            
            key = hashlib.sha256(os.path.splitext(filename)[0].encode()).hexdigest()

            #item = dir_content.get(key)
            if key not in dir_content:
                dir_content[key] = {
                        "path": None,
                        "type": None,
                        "converted": False,
                        "processed": False
                }
            

            if extension in config.extension:
                dir_content[key]["type"] = extension[1:]
                dir_content[key]["path"] = filename
            elif extension == ".md":
                dir_content[key]["converted"] = True
            elif extension == ".json":
                dir_content[key]["processed"] = True
                
        #TODO
        # Se il dir_content[key]["path"] == None significa che per qualche
        # motivo manca il documento e non può essere fatto l'upload. Da gestire
        
        return dir_content

    def cf_stats(self, custom_fields):
        cf_statistics = defaultdict(Counter)

        for custom_field in custom_fields:
            field = custom_field.custom_field_to_dict()
            for key, value in field.items():
                if isinstance(value, (list, set)):  # Unifica list e set
                    cf_statistics[key].update(value)
                else:  # Considera tutto il resto (int, str, ecc.)
                    cf_statistics[key][value] += 1

        for k in cf_statistics:
            sorted = cf_statistics[k].most_common()
            cf_statistics[k] = sorted

        return cf_statistics
    

    def is_acceptable_data(self, date_to_check: str) -> bool:
        """
        Controlla che la data passata come argomento sia un valido formato ISO YYYY-MM-DD
        
        Args:
            date_to_check (str): una stringsa che contiene la data da validare.

        Returns:
            True se la data è nel formato desiderato
            False in ogni altro caso
        
        """
        pattern = r"^\d{4}-\d{2}-\d{2}$"
        return re.match(pattern, date_to_check) is not None


    def get_last_change(self, filename: str) -> Optional[str]:
        """
        Legge la data di modifica di un file

        Args:
            filename (str): la stringa contentente il path del file

        Returns:
            Optional[str]: la data in formato ISO (YYYY-MM-DD) o None in caso di errore

        Example:
            >>> parser = Tagger()
            >>> parser.get_last_change("/mnt/data/file.txt")
            '2023-10-30'    
        """
        try:
            t_modifica = os.path.getmtime(filename)
            return datetime.datetime.fromtimestamp(t_modifica).date().isoformat()
        except FileNotFoundError:
            log.error(f"file {filename} non trovato")
        except Exception as e:
            log.error(f"Errore generico: {e}")
            log.error(traceback.format_exc())
            
        return None


    def parse_json(self, json_string: str) -> Dict[str, Any]:
        """
        Estrae e analizza un oggetto JSON racchiuso tra delimitatori ```

        Args:
            json_string (str): La stringa contenente il JSON formattato con ```json ... ```.
            
        Returns:
            Dict[str, Any]: Il dizionario risultante dal parsing del JSON.
        
        Raises:
            ValueError: Se il JSON non è valido o non viene trovato un blocco JSON corretto.

        Example:
            >>> parser = Tagger()
            >>> parser.parse_json("abcdef ```json {\"key\": \"value\"} ``` ghilmn")
            {'key': 'value'}    
            
        """   
        
        start_index = json_string.find("```json") + len("```json")
        end_index = json_string.find("```", start_index)
        
        if start_index == -1 or end_index == -1:
            raise ValueError("Blocco JSON non trovato o formattazione errata")

        extracted_json = json_string[start_index:end_index].strip()
        
        try:
            data = json.loads(extracted_json)
        except json.JSONDecodeError as e:
            raise ValueError(f"Errore nel parsing del JSON: {e}")

        return data
